/*
 * Copyright (c) 2019 Adrian Siekierka
 *
 * This file is part of Redstone Minus Redstone.
 *
 * Redstone Minus Redstone is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Redstone Minus Redstone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Redstone Minus Redstone.  If not, see <http://www.gnu.org/licenses/>.
 */

package pl.asie.redstoneminus;

import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.Entity;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.SoundCategory;
import net.minecraft.util.SoundEvent;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.ChunkPos;
import net.minecraft.world.IWorldEventListener;
import net.minecraft.world.World;
import net.minecraft.world.chunk.Chunk;

import javax.annotation.Nullable;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public class EnergyWorldTrackerListBased implements IEnergyWorldTracker, IWorldEventListener {
	private final PowerTicker ticker = new PowerTicker();
	private final World world;
	private final Map<ChunkPos, Set<BlockPos>> positions = new HashMap<>();
	private boolean added;

	public EnergyWorldTrackerListBased(World world) {
		this.world = world;
	}

	public void register() {
		if (!added) {
			world.func_72954_a(this);
			added = true;
		}
	}

	public void unregister() {
		if (added) {
			added = false;
			world.func_72848_b(this);
		}
	}

	@Override
	public void chunkLoaded(Chunk chunk) {
		int startX = (chunk.field_76635_g << 4);
		int startZ = (chunk.field_76647_h << 4);
		int endX = startX + 15;
		int endZ = startZ + 15;
		int startY = 0;
		int endY = 255;
		Set<BlockPos> set = new HashSet<>();

		for (BlockPos pos : BlockPos.func_191531_b(startX, startY, startZ, endX, endY, endZ)) {
			IBlockState newState = chunk.func_177435_g(pos);
			if (RedstoneMinusRedstone.blocks.contains(newState.func_177230_c())) {
				set.add(pos.func_185334_h());
			}
		}

		if (set.isEmpty()) {
			positions.remove(chunk.func_76632_l());
		} else {
			positions.put(chunk.func_76632_l(), set);
		}
	}

	@Override
	public void chunkUnloaded(Chunk chunk) {
		positions.remove(chunk.func_76632_l());
	}

	public void tick() {
		int[] ramp = ticker.tickRamp(RedstoneMinusRedstone.powerRamp);
		if (ramp == null) {
			return;
		}

		Set<ChunkPos> deletedPos = new HashSet<>();

		for (Map.Entry<ChunkPos, Set<BlockPos>> entry : positions.entrySet()) {
			ChunkPos chunkPos = entry.getKey();
			Chunk chunk = world.func_72863_F().func_186026_b(chunkPos.field_77276_a, chunkPos.field_77275_b);
			if (chunk == null) {
				deletedPos.add(chunkPos);
				continue;
			}

			for (BlockPos pos : entry.getValue()) {
				TileEntity tile = chunk.func_177424_a(pos, Chunk.EnumCreateEntityType.CHECK);
				if (tile != null) {
					RedstoneMinusHelpers.tick(world, tile, ramp);
				}
			}
		}

		for (ChunkPos pos : deletedPos) {
			positions.remove(pos);
		}
	}

	@Override
	public void func_184376_a(World worldIn, BlockPos pos, IBlockState oldState, IBlockState newState, int flags) {
		ChunkPos chunkPos = new ChunkPos(pos);
		if (RedstoneMinusRedstone.blocks.contains(newState.func_177230_c())) {
			Set<BlockPos> set = positions.computeIfAbsent(chunkPos, (a) -> new HashSet<>());
			set.add(pos);
		} else {
			Set<BlockPos> set;
			if ((set = positions.get(chunkPos)) != null) {
				if (set.remove(pos)) {
					if (set.isEmpty()) {
						positions.remove(chunkPos);
					}
				}
			}
		}
	}

	@Override
	public void func_174959_b(BlockPos pos) {

	}

	@Override
	public void func_147585_a(int x1, int y1, int z1, int x2, int y2, int z2) {

	}

	@Override
	public void func_184375_a(@Nullable EntityPlayer player, SoundEvent soundIn, SoundCategory category, double x, double y, double z, float volume, float pitch) {

	}

	@Override
	public void func_184377_a(SoundEvent soundIn, BlockPos pos) {

	}

	@Override
	public void func_180442_a(int particleID, boolean ignoreRange, double xCoord, double yCoord, double zCoord, double xSpeed, double ySpeed, double zSpeed, int... parameters) {

	}

	@Override
	public void func_190570_a(int id, boolean ignoreRange, boolean minimiseParticleLevel, double x, double y, double z, double xSpeed, double ySpeed, double zSpeed, int... parameters) {

	}

	@Override
	public void func_72703_a(Entity entityIn) {

	}

	@Override
	public void func_72709_b(Entity entityIn) {

	}

	@Override
	public void func_180440_a(int soundID, BlockPos pos, int data) {

	}

	@Override
	public void func_180439_a(EntityPlayer player, int type, BlockPos blockPosIn, int data) {

	}

	@Override
	public void func_180441_b(int breakerId, BlockPos pos, int progress) {

	}
}
