/*
 * Copyright (c) 2019 Adrian Siekierka
 *
 * This file is part of Redstone Minus Redstone.
 *
 * Redstone Minus Redstone is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Redstone Minus Redstone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Redstone Minus Redstone.  If not, see <http://www.gnu.org/licenses/>.
 */

package pl.asie.redstoneminus;

import it.unimi.dsi.fastutil.ints.Int2ObjectMap;
import it.unimi.dsi.fastutil.ints.Int2ObjectOpenHashMap;
import it.unimi.dsi.fastutil.objects.Object2IntMap;
import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;
import net.minecraft.entity.item.EntityItem;
import net.minecraft.item.ItemStack;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.World;
import net.minecraftforge.energy.CapabilityEnergy;
import net.minecraftforge.energy.IEnergyStorage;
import net.minecraftforge.fml.common.eventhandler.SubscribeEvent;
import net.minecraftforge.fml.common.gameevent.TickEvent;
import net.minecraftforge.fml.relauncher.Side;

public class RedstoneInductiveCharger {
    private final Int2ObjectMap<PowerTicker> tickerMap = new Int2ObjectOpenHashMap<>();

    @SubscribeEvent
    public void performInductiveCharging(TickEvent.WorldTickEvent event) {
        if (event.phase == TickEvent.Phase.END && event.side == Side.SERVER) {
            PowerTicker ticker = tickerMap.computeIfAbsent(event.world.field_73011_w.getDimension(), (a) -> new PowerTicker());
            int[] ramp = ticker.tickRamp(RedstoneMinusRedstone.powerRampItemCharging);
            if (ramp == null) {
                return;
            }

            World world = event.world;
            Object2IntMap<BlockPos> posToPower = new Object2IntOpenHashMap<>();
            boolean checkWl = !RedstoneMinusRedstone.inductiveItems.isEmpty();

            event.world.func_175644_a(EntityItem.class, entityItem -> true).forEach((entityItem) -> {
                ItemStack stack = entityItem.func_92059_d();
                if (checkWl && !RedstoneMinusRedstone.inductiveItems.contains(stack.func_77973_b())) {
                    return;
                }

                if (!stack.func_190926_b() && stack.func_190916_E() == 1 && stack.hasCapability(CapabilityEnergy.ENERGY, null)) {
                    if (RedstoneMinusRedstone.preventDespawnInductiveCharging) {
                        entityItem.func_174873_u();
                    }

                    IEnergyStorage storage = stack.getCapability(CapabilityEnergy.ENERGY, null);
                    if (storage != null) {
                        int targetPower = posToPower.computeIfAbsent(entityItem.func_180425_c(), (pos) -> {
                            int power = RedstoneMinusHelpers.getPower(world, pos, EnumFacing.UP);

                            for (EnumFacing facing : EnumFacing.field_82609_l) {
                                if (power >= 15) break;
                                BlockPos posOpp = pos.func_177972_a(facing);
                                power = Math.max(power, RedstoneMinusHelpers.getPower(world, posOpp, facing));
                            }
                            return power;
                        });

                        if (targetPower <= 0) return;
                        targetPower = targetPower > 15 ? ramp[14] : ramp[targetPower - 1];
                        if (targetPower <= 0) return;

                        storage.receiveEnergy(targetPower, false);
                    }
                }
            });
        }
    }
}
