/*
 * Copyright (c) 2019 Adrian Siekierka
 *
 * This file is part of Redstone Minus Redstone.
 *
 * Redstone Minus Redstone is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Redstone Minus Redstone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Redstone Minus Redstone.  If not, see <http://www.gnu.org/licenses/>.
 */

package pl.asie.redstoneminus;

import com.google.common.collect.Sets;
import net.minecraft.block.Block;
import net.minecraft.item.Item;
import net.minecraft.util.ResourceLocation;
import net.minecraft.world.World;
import net.minecraftforge.common.MinecraftForge;
import net.minecraftforge.common.config.Configuration;
import net.minecraftforge.event.world.ChunkEvent;
import net.minecraftforge.event.world.WorldEvent;
import net.minecraftforge.fml.common.Mod;
import net.minecraftforge.fml.common.SidedProxy;
import net.minecraftforge.fml.common.event.FMLInitializationEvent;
import net.minecraftforge.fml.common.event.FMLPostInitializationEvent;
import net.minecraftforge.fml.common.event.FMLPreInitializationEvent;
import net.minecraftforge.fml.common.eventhandler.SubscribeEvent;
import net.minecraftforge.fml.common.gameevent.TickEvent;
import net.minecraftforge.fml.common.registry.ForgeRegistries;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.util.*;

@Mod(
		modid = RedstoneMinusRedstone.MODID,
		name = "Redstone Minus Redstone",
		version = RedstoneMinusRedstone.VERSION
)
public class RedstoneMinusRedstone {
	public static final String MODID = "redstoneminusredstone";
	public static final String VERSION = "@VERSION@";
	public static Logger logger;

	@SidedProxy(serverSide = "pl.asie.redstoneminus.ProxyCommon", clientSide = "pl.asie.redstoneminus.ProxyClient")
	public static ProxyCommon proxy;

	static float[] powerRamp = new float[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 };
	static float[] powerRampItemCharging = new float[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 };
	static Set<Block> blocks = Sets.newIdentityHashSet();
	static Set<Item> inductiveItems = Sets.newIdentityHashSet();
	static boolean disableOptimizations;
	static boolean enableRedstonePower;
	static boolean enableInductiveCharging;
	static boolean preventDespawnInductiveCharging;
	static int tickGrouping = 1;

	private static String[] cfgBlockIds = new String[0];
	private static String[] cfgInductiveItemIds = new String[0];

	private float[] readPowerRamp(String name, Configuration config) {
		float[] ramp = new float[15];
		String[] prStrDef = new String[] {
				"1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12", "13", "14", "15"
		};

		String[] powerRampStr = config.getStringList(name, "balance", prStrDef,
				"The mapping of redstone values {1, 2, ... 15} to energy units. Must have exactly fifteen entries.");

		if (powerRampStr.length != prStrDef.length) {
			powerRampStr = prStrDef;
			config.get("balance", "powerRamp", prStrDef).set(prStrDef);
		}

		for (int i = 0; i < powerRampStr.length; i++) {
			try {
				ramp[i] = Float.parseFloat(powerRampStr[i].trim());
				if (ramp[i] < 0) {
					logger.error("Invalid value @ " + i + ": " + ramp[i] + " < 0!");
					ramp[i] = 0;
				}
			} catch (Exception e) {
				logger.error("Invalid value @ " + i, e);
				ramp[i] = 0;
			}
		}

		return ramp;
	}

	@Mod.EventHandler
	public void onPreInit(FMLPreInitializationEvent event) {
		logger = LogManager.getLogger(MODID);

		Configuration config = new Configuration(event.getSuggestedConfigurationFile());

		powerRamp = readPowerRamp("powerRamp", config);
		powerRampItemCharging = readPowerRamp("powerRampItemCharging", config);

		preventDespawnInductiveCharging = config.getBoolean("preventDespawningItems", "balance", true,
				"If true, items which can be charged 'inductively' will not be despawned.");

		cfgBlockIds = config.getStringList("allowedBlocks", "balance", new String[0],
				"If any valid entries are present, only blocks with these IDs will be powerable with redstone. This also helps performance.");

		cfgInductiveItemIds = config.getStringList("allowedChargeableItems", "balance", new String[0],
				"If any valid entries are present, only items with these IDs will be chargeable with redstone when thrown.");

		disableOptimizations = config.getBoolean("disableOptimizations", "performance", false,
				"Disable this if the mod author tells you to, or if you want to provide extra information for a bug report.");
		tickGrouping = config.getInt("executionTickDivider", "performance", 1, 1, 100,
				"Run the redstone charging code every X ticks, multiplying the power ramp by said X. As such, the total amount of energy is unchanged, but performance costs are reduced.");

		enableRedstonePower = config.getBoolean("redstoneGeneratesForgeEnergy", "features", true,
					"Enable the 'redstone power generates Forge Energy (commonly referred to as RF)' feature.");
		enableInductiveCharging = config.getBoolean("redstoneChargesItemsForgeEnergy", "features", true,
				"Enable the 'redstone charges thrown items containing Forge Energy' feature.");

		if (config.hasChanged()) {
			config.save();
		}
	}

	@Mod.EventHandler
	public void onInit(FMLInitializationEvent event) {
		MinecraftForge.EVENT_BUS.register(this);

		if (enableInductiveCharging) {
			MinecraftForge.EVENT_BUS.register(new RedstoneInductiveCharger());
		}
	}

	@Mod.EventHandler
	public void onPostInit(FMLPostInitializationEvent event) {
		for (String s : cfgBlockIds) {
			Block b = ForgeRegistries.BLOCKS.getValue(new ResourceLocation(s.trim()));
			if (b != null) {
				blocks.add(b);
			}
		}

		for (String s : cfgInductiveItemIds) {
			Item b = ForgeRegistries.ITEMS.getValue(new ResourceLocation(s.trim()));
			if (b != null) {
				inductiveItems.add(b);
			}
		}
	}

	private Map<World, IEnergyWorldTracker> worldTrackers = new IdentityHashMap<>();

	@SubscribeEvent
	public void onLoad(WorldEvent.Load event) {
		if (!enableRedstonePower) {
			return;
		}

		if (!event.getWorld().field_72995_K) {
			if (!worldTrackers.containsKey(event.getWorld())) {
				IEnergyWorldTracker ewt;
				if (blocks.isEmpty() || event.getWorld().func_72800_K() != 256 || disableOptimizations) {
					ewt = new EnergyWorldTrackerEveryone(event.getWorld());
				} else {
					ewt = new EnergyWorldTrackerListBased(event.getWorld());
				}

				ewt.register();

				worldTrackers.put(event.getWorld(), ewt);
			}
		}
	}

	@SubscribeEvent
	public void onLoadChunk(ChunkEvent.Load event) {
		IEnergyWorldTracker ewt = worldTrackers.get(event.getWorld());
		if (ewt != null) {
			ewt.chunkLoaded(event.getChunk());
		}
	}

	@SubscribeEvent
	public void onUnloadChunk(ChunkEvent.Unload event) {
		IEnergyWorldTracker ewt = worldTrackers.get(event.getWorld());
		if (ewt != null) {
			ewt.chunkUnloaded(event.getChunk());
		}
	}

	@SubscribeEvent
	public void onTick(TickEvent.WorldTickEvent event) {
		if (event.phase == TickEvent.Phase.END) {
			IEnergyWorldTracker ewt = worldTrackers.get(event.world);
			if (ewt != null) {
				ewt.tick();
			}
		}
	}

	@SubscribeEvent
	public void onUnload(WorldEvent.Unload event) {
		if (worldTrackers.containsKey(event.getWorld())) {
			IEnergyWorldTracker ewt = worldTrackers.remove(event.getWorld());
			ewt.unregister();
		}
	}
}
